﻿using System;
using System.Linq;
using System.Windows.Forms;
using System.IO.Ports;          // added to include serial ports
using System.IO;
using System.Threading;

namespace SRF485_test
{
    public partial class Form1 : Form
    {
        MainMenu MyMenu;                            // declares a top menubar (file etc)
        string comportname;                         // a string to contain used port "COM3" for example
        static SerialPort USB_RS485;                // declares a serial port handle
        string[] comportnames = new string[100];    // a string array that I fill will all available com ports at the program start
        byte[] buffer = new byte[10];               // the I/O buffer for send/recieve to the USB-RS485
        byte srf_485_found = 0;                     // a flag to indicate search routine has found a srf485
        int srf485_address;                         // the found srf485 address 
 
        enum commands   // the srf485 command set, where xx commands are unused 
        {
            REAL_IN = 0x50, REAL_CM, REAL_US, xx7, xx8, xx9, FAKE_IN, FAKE_CM, FAKE_US, xx10, xx11, xx12,
            BURST, GET_VER, GET_RANGE, xx1, xx2, xx3, xx4, xx5, xx6, SET_SEARCH, LESS_THAN, SET_GROUP,
            GET_TEMP, GET_TRANGE, xx18, xx19, xx20, xx21, TESTRANGE,
        };

        public Form1()
        {

            string last_used_comport;
            int x;

            InitializeComponent();
            USB_RS485 = new SerialPort();
            MyMenu = new MainMenu();                
            MenuItem m1 = new MenuItem("File");         // declare file menu and submenu exit
            MenuItem subm1 = new MenuItem("Exit");
            m1.MenuItems.Add(subm1);

            MenuItem m2 = new MenuItem("Setup");
            MenuItem subm2 = new MenuItem("Set Comport");
            m2.MenuItems.Add(subm2);

            Menu = MyMenu;
            MyMenu.MenuItems.Add(m1);       // add the declared menu items to the menu
            MyMenu.MenuItems.Add(m2);

            subm1.Click += new EventHandler(MMExitClick);   // declares an event, when exit is clicked function MMExitClick will be called

            try { last_used_comport = Application.UserAppDataRegistry.GetValue("last_comport").ToString(); } // try to look at registry for last used comport
            catch
            {
                last_used_comport = null;   // otherwise set it to null
            }

            comportnames = SerialPort.GetPortNames();   // gets all available comports names into string array
            var sortedList = comportnames.OrderBy(port => Convert.ToInt32(port.Replace("COM", string.Empty)));  // sorts comports into ascending order
            x = 0;
            foreach (string port in sortedList) // loop adds available comports into menu bar, and opens comport if registry gave a last used port value
            {
                subm2.MenuItems.Add(port);
                subm2.MenuItems[x].Click += new EventHandler(MMCommClick);  // event handler for press of comx in menu bar
                if (port == last_used_comport)
                {
                    MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = true;
                    comportname = port;
                    OpenCommPort();
                }
                x++;
            }
        }

        protected void MMCommClick(object who, EventArgs e) // called when comx is pressed in menu bar
        {
            byte x;
            string senderinfo;
            senderinfo = who.ToString();
            comportname = senderinfo.Substring(53);     // rips comx menu bar item that was pressed from string who
            for (x = 0; x < comportnames.Length; x++)   // loops round checking or unchecking comx
            {
                if (comportnames[x] == comportname) MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = true;
                else if (comportnames[x] != null) MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = false;
            }
            OpenCommPort();
            Application.UserAppDataRegistry.SetValue("last_comport", comportname);          // sets last_comport so next time program will try it first
        }

        private void OpenCommPort()
        {
            byte x;
            if (USB_RS485.IsOpen) USB_RS485.Close();
            USB_RS485.PortName = (string)comportname;    // retrieves "COMx"
            USB_RS485.Parity = 0;
            USB_RS485.BaudRate = 19200;
            USB_RS485.StopBits = StopBits.Two;
            USB_RS485.DataBits = 8;
            USB_RS485.ReadTimeout = 500;
            USB_RS485.WriteTimeout = 500;
            USB_RS485.Handshake = Handshake.None;
            try
            {
                USB_RS485.Open();
                GC.SuppressFinalize(USB_RS485.BaseStream);
            }
            catch
            {
                MessageBox.Show("error opening comport");
                for (x = 0; x < 100; x++)
                {
                    if (comportnames[x] != null) MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = false;
                }
            }
            if (USB_RS485.IsOpen)
            {
                label1.Text = "Searching for SRF485";
                timer1.Enabled = true;
                timer1.Start();
            }
        }

        private void get_range_and_temp()
        {
            int range;
            int temp;

            buffer[0] = 0;
            USB_RS485.BaudRate = 19200;                      // sending 0x00 at 19200 is a break at 38400
            rs485_write(1);
 
            USB_RS485.BaudRate = 38400;

            buffer[0] = (byte)commands.GET_TRANGE;
            buffer[1] = (byte)((srf485_address >> 16) & 255);	                	    	// AddrH	
            buffer[2] = (byte)((srf485_address >> 8) & 255);	                	    	// AddrM
            buffer[3] = (byte)(srf485_address & 255);		                			// AddrL
            buffer[4] = 0x00;					                            	// Data
            buffer[5] = (byte)(~(buffer[0] + buffer[1] + buffer[2] + buffer[3] + buffer[4]));	// Chksum
            rs485_write(6);
            rs485_recieve(2);
            range = (buffer[0] << 8);
            range += buffer[1];
            textBox2.Text = string.Format("{0}cm", range);

            buffer[0] = 0;
            USB_RS485.BaudRate = 19200;                      // sending 0x00 at 19200 is a break at 38400
            rs485_write(1);
            
            USB_RS485.BaudRate = 38400;

            buffer[0] = (byte)commands.GET_TEMP;
            buffer[1] = (byte)((srf485_address >> 16) & 255);	                	    	// AddrH	
            buffer[2] = (byte)((srf485_address >> 8) & 255);	                	    	// AddrM
            buffer[3] = (byte)(srf485_address & 255);		                			// AddrL
            buffer[4] = 0x00;					                            	// Data
            buffer[5] = (byte)(~(buffer[0] + buffer[1] + buffer[2] + buffer[3] + buffer[4]));	// Chksum
            rs485_write(6);
            rs485_recieve(2);
            temp = (buffer[0] << 8);
            temp += buffer[1];
            textBox3.Text = string.Format("{0}°C", temp);
            
            buffer[0] = 0;
            USB_RS485.BaudRate = 19200;                      // sending 0x00 at 19200 is a break at 38400
            rs485_write(1);
            USB_RS485.BaudRate = 38400;
            buffer[0] = (byte)commands.REAL_CM;
            buffer[1] = (byte)((srf485_address >> 16) & 255);	                	    	// AddrH	
            buffer[2] = (byte)((srf485_address >> 8) & 255);	                	    	// AddrM
            buffer[3] = (byte)(srf485_address & 255);		                			// AddrL
            buffer[4] = 0x00;					                            	// Data
            buffer[5] = (byte)(~(buffer[0] + buffer[1] + buffer[2] + buffer[3] + buffer[4]));	// Chksum
            rs485_write(6);
        }

        private void rs485_write(byte buffer_count)
        {
            byte x;

            if (USB_RS485.IsOpen)
            {
                try
                {
                    USB_RS485.Write(buffer, 0, buffer_count);
                    while (USB_RS485.BytesToWrite > 0) ;
                    Thread.Sleep(5);
                }
                catch
                {
                    timer1.Enabled = false;
                    try
                    {
                        GC.ReRegisterForFinalize(USB_RS485.BaseStream);
                        USB_RS485.Dispose();
                    }
                    catch { }
                    MessageBox.Show("Communication error, comport closed");
                    label1.Text = "Waiting for COM port selection";
                    srf_485_found = 0;
                    for (x = 0; x < comportnames.Length; x++)
                    {
                        if (comportnames[x] != null) MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = false;
                    }
                    
                }
            }
        }

        private void rs485_recieve(byte read_bytes)
        {
            byte x, message_shown = 0;

            for (x = 0; x < read_bytes && USB_RS485.IsOpen; x++)       // this will call the read function for the passed number times, 
            {                                      // this way it ensures each byte has been correctly recieved while
                try                                // still using timeouts
                {
                    USB_RS485.Read(buffer, x, 1);     // retrieves 1 byte at a time and places in SerBuf at position x
                }
                catch
                {
                    timer1.Enabled = false; 
                    if (message_shown == 0) MessageBox.Show(string.Format("Communication error, comport closed - read {0}/{1}", x, read_bytes));
                    label1.Text = "Waiting for COM port selection";
                    srf_485_found = 0;
                    message_shown = 1;
                    try
                    {
                        GC.ReRegisterForFinalize(USB_RS485.BaseStream);
                        USB_RS485.Dispose();
                    }
                    catch { }
                    for (x = 0; x < comportnames.Length; x++)
                    {
                        if (comportnames[x] != null) MyMenu.MenuItems[1].MenuItems[0].MenuItems[x].Checked = false;
                    }
                }
            }
        }

        private void SetSearchMode()
        {
            buffer[0] = 0;
            USB_RS485.BaudRate = 19200;
            rs485_write(1);
            USB_RS485.BaudRate = 38400;
            buffer[0] = (byte) commands.SET_SEARCH;
            buffer[1] = 0;	                	    	                    // AddrH	
            buffer[2] = 0;	                                       	    	// AddrM
            buffer[3] = 0;                                           			// AddrL
            buffer[4] = 0;					                            	// Data
            buffer[5] = (byte) (~(buffer[0] + buffer[1] + buffer[2] + buffer[3] + buffer[4]));	// Chksum
            rs485_write(6);
            System.Threading.Thread.Sleep(1);
        }

        // Be sure to have placed SRF485's in search mode (SET_SEARCH) before calling this function
        private void FindModule()
        {
            int Walk;
            Walk = 0x800000;
            srf485_address = 0x800000;

            do
            {
                buffer[0] = 0;
                USB_RS485.BaudRate = 19200;                      // sending 0x00 at 19200 is a break at 38400
                rs485_write(1);
                USB_RS485.BaudRate = 38400;
                
                buffer[0] = (byte)commands.LESS_THAN;
                buffer[1] = (byte)((srf485_address >> 16) & 255);	                	    	// AddrH	
                buffer[2] = (byte)((srf485_address >> 8) & 255);	                	    	// AddrM
                buffer[3] = (byte)(srf485_address & 255);		                			// AddrL
                buffer[4] = 0x00;					                            	// Data
                buffer[5] = (byte)(~(buffer[0] + buffer[1] + buffer[2] + buffer[3] + buffer[4]));	// Chksum
                rs485_write(6);
                System.Threading.Thread.Sleep(21); //16ms of port latency time (default for FTDI) added to 5ms for response

                if (USB_RS485.BytesToRead > 0)
                {
                    srf485_address ^= Walk;
                    rs485_recieve((byte) USB_RS485.BytesToRead);
                }
                Walk >>= 1;
                srf485_address |= Walk;
                textBox1.Text = string.Format("{0:X06}", srf485_address);
                textBox1.Update();
            }
            while (Walk > 0);
            srf_485_found = 1;
        }
         
        // Handler for main menu Exit selection. 
        protected void MMExitClick(object who, EventArgs e)
        {
            Application.Exit();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            timer1.Stop();
            if (USB_RS485.IsOpen)
            {
                if (srf_485_found == 0)
                {
                    SetSearchMode();
                    FindModule();
                }
                else
                {
                    label1.Text = "Measuring range and temp";
                    get_range_and_temp();
                }
            }
            timer1.Start();
        }

        private void textBox_Enter(object sender, EventArgs e)  // stops user being able to put cursor in textbox, just selects label1 at that point
        {
            label1.Select();
        }

    }
}
